<?php

namespace Drupal\vipps_mobilepay_commerce\Controller;

use CommerceGuys\Intl\Formatter\CurrencyFormatterInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use zaporylie\Vipps\Api\v1\EPayment as EPaymentUpstream;
use zaporylie\Vipps\Model\EPayment\v1\Aggregate;

/**
 * Returns responses for Commerce Vipps routes.
 */
class TransactionLogController extends ControllerBase {

  /**
   * The vipps manager factory.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface
   */
  protected $vippsManagerFactory;

  /**
   * The commerce_price.currency_formatter service.
   *
   * @var \CommerceGuys\Intl\Formatter\CurrencyFormatterInterface
   */
  protected $commercePriceCurrencyFormatter;

  /**
   * The minor units converter.
   *
   * @var \Drupal\commerce_price\MinorUnitsConverterInterface
   */
  protected $minorUnitsConverter;

  /**
   * The controller constructor.
   *
   * @param \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface $vipps_mobilepay_factory
   *   The vipps factory.
   * @param \CommerceGuys\Intl\Formatter\CurrencyFormatterInterface $commerce_price_currency_formatter
   *   The currency formatter.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $minor_units_converter
   *   The minor units converter.
   */
  public function __construct(VippsMobilePayManagerFactoryInterface $vipps_mobilepay_factory, CurrencyFormatterInterface $commerce_price_currency_formatter, MinorUnitsConverterInterface $minor_units_converter) {
    $this->vippsManagerFactory = $vipps_mobilepay_factory;
    $this->commercePriceCurrencyFormatter = $commerce_price_currency_formatter;
    $this->minorUnitsConverter = $minor_units_converter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('vipps_mobilepay.manager_factory'),
      $container->get('commerce_price.currency_formatter'),
      $container->get('commerce_price.minor_units_converter')
    );
  }

  /**
   * Builds the response.
   */
  public function build(PaymentInterface $payment) {

    $plugin = $payment->getPaymentGateway()->getPlugin();
    $settings = $payment->getPaymentGateway()->getPluginConfiguration();

    $connection = $this->entityTypeManager()->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $vipps = $this->vippsManagerFactory->get($connection);
    $payment_manager = new EPaymentUpstream($vipps);
    $response = $payment_manager->getPayment($payment->getRemoteId());

    // Transaction summary.
    if ($response->getAggregate() instanceof Aggregate) {
      $aggregate = $response->getAggregate();
      $rows = [
        [
          ['data' => $this->t('Captured Amount'), 'header' => TRUE],
          $this->commercePriceCurrencyFormatter->format($this->minorUnitsConverter->fromMinorUnits($aggregate->getCapturedAmount()->getValue(), $aggregate->getCapturedAmount()->getCurrency())->getNumber(), $aggregate->getCapturedAmount()->getCurrency()),
        ],
        [
          ['data' => $this->t('Refunded Amount'), 'header' => TRUE],
          $this->commercePriceCurrencyFormatter->format($this->minorUnitsConverter->fromMinorUnits($aggregate->getRefundedAmount()->getValue(), $aggregate->getRefundedAmount()->getCurrency())->getNumber(), $aggregate->getRefundedAmount()->getCurrency()),
        ],
        [
          ['data' => $this->t('Authorized Amount'), 'header' => TRUE],
          $this->commercePriceCurrencyFormatter->format($this->minorUnitsConverter->fromMinorUnits($aggregate->getAuthorizedAmount()->getValue(), $aggregate->getAuthorizedAmount()->getCurrency())->getNumber(), $aggregate->getAuthorizedAmount()->getCurrency()),
        ],
        [
          ['data' => $this->t('Cancelled Amount'), 'header' => TRUE],
          $this->commercePriceCurrencyFormatter->format($this->minorUnitsConverter->fromMinorUnits($aggregate->getCancelledAmount()->getValue(), $aggregate->getCancelledAmount()->getCurrency())->getNumber(), $aggregate->getCancelledAmount()->getCurrency()),
        ],
      ];
      $build['transaction_summary'] = [
        '#type' => 'table',
        '#rows' => $rows,
        '#caption' => $this->t('Transaction Summary'),
      ];
    }

    $response = $payment_manager->getPaymentEvents($payment->getRemoteId());

    // Transaction Log.
    if (is_array($response)) {
      $header = [
        $this->t('Operation'),
        $this->t('Success'),
        $this->t('Amount'),
        $this->t('Timestamp'),
        $this->t('PSP Reference'),
      ];
      $rows = [];
      /** @var \zaporylie\Vipps\Model\EPayment\v1\EventLog $logEntry */
      foreach ($response as $logEntry) {
        $rows[] = [
          $logEntry->getName(),
          $logEntry->getSuccess() ? $this->t('Yes') : $this->t('No'),
          $this->commercePriceCurrencyFormatter->format($this->minorUnitsConverter->fromMinorUnits($logEntry->getAmount()->getValue(), $logEntry->getAmount()->getCurrency())->getNumber(), $logEntry->getAmount()->getCurrency()),
          $logEntry->getTimeStamp()->format('c'),
          $logEntry->getPspReference(),
        ];
      }
      $build['transaction_log'] = [
        '#type' => 'table',
        '#header' => $header,
        '#rows' => $rows,
        '#caption' => $this->t('Transaction Log'),
      ];
    }

    return $build;
  }

}
