<?php

namespace Drupal\vipps_mobilepay_commerce\Form;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\CurrentRouteMatch;
use Drupal\Core\Url;
use Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use zaporylie\Vipps\Api\v1\Webhook as WebhookSDK;
use zaporylie\Vipps\Model\Webhook\v1\RegisterWebhookRequest;

/**
 * The webhook form.
 */
class Webhook extends FormBase {

  const EVENT_NAME = 'epayments.payment.authorized.v1';

  /**
   * The payment gateway.
   *
   * @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface
   */
  protected $paymentGateway;

  /**
   * The Vipps Manager.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface
   */
  protected $managerFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new OrderReassignForm object.
   *
   * @param \Drupal\Core\Routing\CurrentRouteMatch $current_route_match
   *   The current route match.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface $vipps_mobilepay_manager_factory
   *   The vipps manager factory.
   */
  public function __construct(CurrentRouteMatch $current_route_match, EntityTypeManagerInterface $entity_type_manager, VippsMobilePayManagerFactoryInterface $vipps_mobilepay_manager_factory) {
    $this->paymentGateway = $current_route_match->getParameter('commerce_payment_gateway');
    $this->entityTypeManager = $entity_type_manager;
    $this->managerFactory = $vipps_mobilepay_manager_factory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('current_route_match'),
      $container->get('entity_type.manager'),
      $container->get('vipps_mobilepay.manager_factory'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'vipps_mobilepay_commerce_webhook';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    /** @var \Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway\EPayment $plugin */
    $plugin = $this->paymentGateway->getPlugin();
    $settings = $plugin->getConfiguration();
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $webhook_url = Url::fromRoute('commerce_payment.notify', [
      'commerce_payment_gateway' => $this->paymentGateway->id(),
    ])->setAbsolute()->toString();
    try {
      /** @var \Drupal\vipps_mobilepay\VippsMobilepayConnectionInterface $connection */
      $vipps = $this->managerFactory->get($connection);
      $webhook = new WebhookSDK($vipps);
      $webhooks = $webhook->getWebhooks();
      foreach ($webhooks->getWebhooks() as $webhook) {
        if ($webhook->getUrl() !== $webhook_url) {
          continue;
        }
        $events = array_filter($webhook->getEvents(), function ($event) {
          return $event === Webhook::EVENT_NAME;
        });
        if (!count($events)) {
          continue;
        }
        $form[$webhook->getId()] = [
          '#tree' => TRUE,
          '#type' => 'details',
          '#title' => $this->t('Webhook @id (@url): @events', [
            '@id' => $webhook->getId(),
            '@url' => $webhook->getUrl(),
            '@events' => implode(', ', $webhook->getEvents()),
          ]),
        ];
        $form[$webhook->getId()]['delete'] = [
          '#type' => 'submit',
          '#value' => $this->t('Delete webhook'),
        ];
      }
    }
    catch (\Exception $exception) {
      $this->messenger()->addError('Unable to establish connection with API. Check connection details.');
      $this->logger('vipps_mobilepay_commerce')->error('Unable to connect with Vipps MobilePay API: :message', [':message' => $exception->getMessage()]);
      return $form;
    }

    $form['webhook_url'] = [
      '#type' => 'value',
      '#value' => $webhook_url,
    ];

    $form['create'] = [
      '#type' => 'submit',
      '#value' => $this->t('Register new webhook'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#parents'];
    $trigger = end($parents);

    /** @var \Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway\EPayment $plugin */
    $plugin = $this->paymentGateway->getPlugin();
    $settings = $plugin->getConfiguration();
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    /** @var \Drupal\vipps_mobilepay\VippsMobilepayConnectionInterface $connection */
    $vipps = $this->managerFactory->get($connection);
    $webhook = new WebhookSDK($vipps);

    try {
      if ($trigger == 'create') {
        $request = new RegisterWebhookRequest();
        $request->setUrl($form_state->getValue('webhook_url'));
        $request->setEvents([self::EVENT_NAME]);
        $webhook->registerWebhook($request);
        $this->messenger()
          ->addMessage($this->t('Registered webhook for the URK %url and event %event',
            [
              '%url' => $form_state->getValue('webhook_url'),
              '%event' => 'epayments.payment.authorized.v1',
            ]));
      }
      elseif (($trigger == 'delete') && ($reference = prev($parents))) {
        $webhook->deleteWebhook($reference);
        $this->messenger()
          ->addMessage($this->t('The webhook was successfully deleted'));
      }
    }
    catch (\Exception $exception) {
      $this->messenger()
        ->addError($this->t('Error when performing the operation: %message', ['%message' => $exception->getMessage()]));
    }
  }

}
