<?php

namespace Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsNotificationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\CheckoutLandingPageRedirectForm;
use Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\EPaymentLandingPageRedirectForm;
use Symfony\Component\HttpFoundation\Request;
use zaporylie\Vipps\Api\v3\Checkout as CheckoutApi;

/**
 * Provides the Vipps MobilePay payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "vipps_mobilepay_checkout",
 *   label = "Vipps MobilePay Checkout",
 *   display_label = "Vipps MobilePay Checkout",
 *    forms = {
 *     "offsite-payment" = "Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\CheckoutLandingPageRedirectForm",
 *   },
 *   requires_billing_information = FALSE,
 *   js_library="vipps_mobilepay_commerce/checkout_iframe",
 * )
 */
class Checkout extends EPayment implements SupportsAuthorizationsInterface, SupportsRefundsInterface, SupportsNotificationsInterface {

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $reference = $request->query->get(EPaymentLandingPageRedirectForm::QUERY_NAME);
    $paymentGatewayId = $this->parentEntity->id();
    $lockId = $paymentGatewayId . '__' . $reference;

    // Keep checking if lock could be acquired for 5 seconds, then start over.
    while ($this->lock->wait($lockId, 5) || !$this->lock->acquire($lockId)) {
      // Looks like lock cannot be acquired, hold for one second and retry.
      sleep(1);
      $this->logger->notice('Waiting for lock @lock to be released', ['@lock' => $lockId]);
    }
    $this->logger->notice('Lock @lock was acquired', ['@lock' => $lockId]);

    $matching_payments = $this->entityTypeManager->getStorage('commerce_payment')->loadByProperties([
      'remote_id' => $reference,
      'order_id' => $order->id(),
    ]);
    if (count($matching_payments) > 0) {
      $this->logger->debug('Payment :reference found already, skipping.', [':reference' => $reference]);
      return;
    }

    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $settings = $this->configuration;
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $client = $this->vippsManagerFactory->get($connection);
    $response = (new CheckoutApi($client))->getCheckoutSession($reference);

    $state = $response->getPaymentDetails()->getState();
    switch ($state) {
      case 'AUTHORIZED':
        // Payment was successfully authorized.
        break;

      case 'TERMINATED':
        throw new PaymentGatewayException(sprintf('Payment failed with the status %s', $state));

      // Status CREATED means that final status has not yet been issued, hence
      // order must be kept locked in payment return state and status check must
      // be retried every 5s.
      case 'CREATED':
      default:
        sleep(5);
        $this->logger->error('Incorrect status @status, waiting for status change', ['@status' => $state]);
        throw new NeedsRedirectException(Url::fromRoute('<current>')->toString());
    }

    $payment_amount = $this->minorUnitsConverter->fromMinorUnits($response->getPaymentDetails()->getAmount()->getValue(), $response->getPaymentDetails()->getAmount()->getCurrency());
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Use the amount we get from remote, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $this->parentEntity->id(),
      'order_id' => $order->id(),
      'test' => $this->getMode() === 'test',
      'remote_id' => $response->getReference(),
      'remote_state' => $response->getSessionState(),
    ]);
    $payment->save();
  }

  /**
   * Vipps treats onReturn and onCancel in the same way.
   *
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    static::onReturn($order, $request);
  }

  /**
   * {@inheritdoc}
   */
  public function onNotify(Request $request) {
    $content = $request->getContent();
    $json = Json::decode($content);

    $reference = $json['reference'] ?? NULL;
    $order_reference = $request->query->get(CheckoutLandingPageRedirectForm::QUERY_NAME_ORDER_REFERENCE);

    if (!isset($reference)) {
      $this->logger->error('Payment reference is unknown. Unable to handle the webhook');
      throw new PaymentGatewayException(sprintf('Payment reference is unknown. Unable to handle the webhook.'));
    }

    $paymentGatewayId = $this->parentEntity->id();
    $lockId = $paymentGatewayId . '__' . $reference;

    // Keep checking if lock could be acquired for 5 seconds, then start over.
    while ($this->lock->wait($lockId, 5) || !$this->lock->acquire($lockId)) {
      // Looks like lock cannot be acquired, hold for one second and retry.
      sleep(1);
      $this->logger->notice('Waiting for lock @lock to be released', ['@lock' => $lockId]);
    }
    $this->logger->notice('Lock @lock was acquired', ['@lock' => $lockId]);

    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $settings = $this->configuration;
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $vipps = $this->vippsManagerFactory->get($connection);
    $response = (new CheckoutApi($vipps))->getCheckoutSession($reference);
    if (!isset($order_reference)) {
      $this->logger->error('Order reference is unknown. Unable to handle the webhook');
      throw new PaymentGatewayException(sprintf('Order reference is unknown. Unable to handle the webhook.'));
    }

    $matching_payments = $this->entityTypeManager->getStorage('commerce_payment')->loadByProperties([
      'remote_id' => $reference,
      'order_id' => $order_reference,
    ]);
    if (count($matching_payments) > 0) {
      $this->logger->debug('Payment :reference found already, skipping.', [':reference' => $reference]);
      return;
    }

    // Validate status.
    $state = $response->getPaymentDetails()->getState();
    switch ($state) {
      case 'AUTHORIZED':
        // Payment was successfully authorized.
        break;

      default:
        $this->logger->notice('Webhook handling failed for payment reference :reference with the status :status', [
          ':reference' => $reference,
          ':status' => $state,
        ]);
        return '';
    }

    $payment_amount = $this->minorUnitsConverter->fromMinorUnits($response->getPaymentDetails()->getAmount()->getValue(), $response->getPaymentDetails()->getAmount()->getCurrency());
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Use the amount we get from remote, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $paymentGatewayId,
      'order_id' => $order_reference,
      'test' => $this->getMode() === 'test',
      'remote_id' => $reference,
      'remote_state' => $response->getSessionState(),
    ]);
    $payment->save();
  }

}
