<?php

namespace Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsNotificationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_price\Price;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Url;
use Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\EPaymentLandingPageRedirectForm;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use zaporylie\Vipps\Api\v1\EPayment as EPaymentUpstream;
use zaporylie\Vipps\Exceptions\VippsException;
use zaporylie\Vipps\Model\EPayment\v1\Amount;
use zaporylie\Vipps\Model\EPayment\v1\CancelModificationRequest;
use zaporylie\Vipps\Model\EPayment\v1\CaptureModificationRequest;
use zaporylie\Vipps\Model\EPayment\v1\RefundModificationRequest;

/**
 * Provides the Vipps MobilePay payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "vipps_mobilepay_epayment",
 *   label = "Vipps MobilePay ePayment",
 *   display_label = "Vipps MobilePay ePayment",
 *   forms = {
 *     "offsite-payment" = "Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\EPaymentLandingPageRedirectForm",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class EPayment extends VippsMobilePayPaymentGatewayBase implements SupportsAuthorizationsInterface, SupportsRefundsInterface, SupportsNotificationsInterface {

  /**
   * Vipps manager factory.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface
   */
  protected $vippsManagerFactory;

  /**
   * The lock service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $object = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $object->vippsManagerFactory = $container->get('vipps_mobilepay.manager_factory');
    $object->lock = $container->get('lock');
    $object->logger = $container->get('logger.channel.vipps_mobilepay_commerce');
    return $object;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $reference = $request->query->get(EPaymentLandingPageRedirectForm::QUERY_NAME);
    $paymentGatewayId = $this->parentEntity->id();
    $lockId = $paymentGatewayId . '__' . $reference;

    // Keep checking if lock could be acquired for 5 seconds, then start over.
    while ($this->lock->wait($lockId, 5) || !$this->lock->acquire($lockId)) {
      // Looks like lock cannot be acquired, hold for one second and retry.
      sleep(1);
      $this->logger->notice('Waiting for lock @lock to be released', ['@lock' => $lockId]);
    }
    $this->logger->notice('Lock @lock was acquired', ['@lock' => $lockId]);

    $matching_payments = $this->entityTypeManager->getStorage('commerce_payment')->loadByProperties([
      'remote_id' => $reference,
      'order_id' => $order->id(),
    ]);
    if (count($matching_payments) > 0) {
      $this->logger->debug('Payment :reference found already, skipping.', [':reference' => $reference]);
      return;
    }

    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $settings = $this->configuration;
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $vipps = $this->vippsManagerFactory->get($connection);
    $response = (new EPaymentUpstream($vipps))->getPayment($reference);

    // @todo Validate status.
    $state = $response->getState();
    switch ($state) {
      case 'AUTHORIZED':
        // Payment was successfully authorized.
        break;

      case 'ABORTED':
      case 'EXPIRED':
      case 'TERMINATED':
        throw new PaymentGatewayException(sprintf('Payment failed with the status %s', $state));

      // Status CREATED means that final status has not yet been issued, hence
      // order must be kept locked in payment return state and status check must
      // be retried every 5s.
      case 'CREATED':
      default:
        sleep(5);
        $this->logger->error('Incorrect status @status, waiting for status change', ['@status' => $state]);
        throw new NeedsRedirectException(Url::fromRoute('<current>')->toString());
    }

    $payment_amount = $this->minorUnitsConverter->fromMinorUnits($response->getAmount()->getValue(), $response->getAmount()->getCurrency());
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Use the amount we get from remote, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $this->parentEntity->id(),
      'order_id' => $order->id(),
      'test' => $this->getMode() === 'test',
      'remote_id' => $response->getReference(),
      'remote_state' => $response->getState(),
    ]);
    $payment->save();
  }

  /**
   * Vipps treats onReturn and onCancel in the same way.
   *
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    static::onReturn($order, $request);
  }

  /**
   * {@inheritdoc}
   */
  public function onNotify(Request $request) {
    $content = $request->getContent();
    $json = Json::decode($content);

    $reference = $json['reference'] ?? NULL;
    if (!isset($reference)) {
      throw new PaymentGatewayException(sprintf('Payment reference is unknown. Unable to handle the webhook.'));
    }

    $paymentGatewayId = $this->parentEntity->id();
    $lockId = $paymentGatewayId . '__' . $reference;

    // Keep checking if lock could be acquired for 5 seconds, then start over.
    while ($this->lock->wait($lockId, 5) || !$this->lock->acquire($lockId)) {
      // Looks like lock cannot be acquired, hold for one second and retry.
      sleep(1);
      $this->logger->notice('Waiting for lock @lock to be released', ['@lock' => $lockId]);
    }
    $this->logger->notice('Lock @lock was acquired', ['@lock' => $lockId]);

    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $settings = $this->configuration;
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $vipps = $this->vippsManagerFactory->get($connection);
    $response = (new EPaymentUpstream($vipps))->getPayment($reference);
    $metadata = $response->getMetadata();
    if (!isset($metadata['order_id'])) {
      throw new PaymentGatewayException(sprintf('Order reference is unknown. Unable to handle the webhook.'));
    }
    $order_id = $metadata['order_id'];

    $matching_payments = $this->entityTypeManager->getStorage('commerce_payment')->loadByProperties([
      'remote_id' => $reference,
      'order_id' => $order_id,
    ]);
    if (count($matching_payments) > 0) {
      $this->logger->debug('Payment :reference found already, skipping.', [':reference' => $reference]);
      return;
    }

    // Validate status.
    $state = $response->getState();
    switch ($state) {
      case 'AUTHORIZED':
        // Payment was successfully authorized.
        break;

      case 'ABORTED':
      case 'EXPIRED':
      case 'TERMINATED':
      case 'CREATED':
      default:
        throw new PaymentGatewayException(sprintf('Webhook handling failed for payment reference %s with the status %s', $reference, $state));
    }

    $payment_amount = $this->minorUnitsConverter->fromMinorUnits($response->getAmount()->getValue(), $response->getAmount()->getCurrency());
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'state' => 'authorization',
      // Use the amount we get from remote, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $paymentGatewayId,
      'order_id' => $order_id,
      'test' => $this->getMode() === 'test',
      'remote_id' => $reference,
      'remote_state' => $response->getState(),
    ]);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    // Assert things.
    $this->assertPaymentState($payment, ['authorization']);
    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();

    // Adds support for partial captures.
    if ($amount->lessThan($payment->getAmount())) {
      /** @var \Drupal\commerce_payment\Entity\PaymentInterface $parent_payment */
      $parent_payment = $payment;
      $payment = $parent_payment->createDuplicate();
    }

    $remote_id = $payment->getRemoteId();
    $number = $this->minorUnitsConverter->toMinorUnits($amount);
    $request = new CaptureModificationRequest();
    $modification_amount = (new Amount())
      ->setValue($number)
      ->setCurrency($payment->getAmount()->getCurrencyCode());
    $request->setModificationAmount($modification_amount);
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new EPaymentUpstream($vipps);
      $response = $payment_manager->capturePayment($remote_id, $request, NULL);
    }
    catch (VippsException $exception) {
      throw new DeclineException($exception->getMessage());
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $payment->setState('completed');
    $payment->setAmount($amount);
    $payment->save();

    // Update parent payment if one exists.
    if (isset($parent_payment)) {
      $parent_payment->setAmount($parent_payment->getAmount()->subtract($amount));
      if ($parent_payment->getAmount()->isZero()) {
        $parent_payment->setState('authorization_voided');
      }
      $parent_payment->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);
    $remote_id = $payment->getRemoteId();

    $request = new CancelModificationRequest();
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new EPaymentUpstream($vipps);
      $response = $payment_manager->cancelPayment($remote_id, $request, NULL);
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $payment->setState('authorization_voided');
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    // Validate.
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);

    // Let's do some refunds.
    parent::assertRefundAmount($payment, $amount);

    $remote_id = $payment->getRemoteId();
    $number = $this->minorUnitsConverter->toMinorUnits($amount);
    $request = new RefundModificationRequest();
    $modification_amount = (new Amount())
      ->setValue($number)
      ->setCurrency($payment->getAmount()->getCurrencyCode());
    $request->setModificationAmount($modification_amount);
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new EPaymentUpstream($vipps);
      $response = $payment_manager->refundPayment($remote_id, $request, NULL);
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();

  }

}
