<?php

namespace Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Exception\SoftDeclineException;
use Drupal\commerce_payment\PaymentMethodStorageInterface;
use Drupal\commerce_payment\PaymentStorageInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsStoredPaymentMethodsInterface;
use Drupal\commerce_price\Price;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use zaporylie\Vipps\Api\UserInfo;
use zaporylie\Vipps\Api\v3\RecurringPayment;
use zaporylie\Vipps\Exceptions\VippsException;
use zaporylie\Vipps\Model\RecurringPayment\v3\RequestCaptureCharge;
use zaporylie\Vipps\Model\RecurringPayment\v3\RequestCreateCharge;
use zaporylie\Vipps\Model\RecurringPayment\v3\RequestRefundCharge;
use zaporylie\Vipps\Model\RecurringPayment\v3\RequestUpdateAgreement;

/**
 * Provides the Vipps MobilePay stored payment method gateway.
 *
 * @CommercePaymentGateway(
 *   id = "vipps_mobilepay_recurring_payment_method",
 *   label = "Vipps MobilePay Recurring Payment Method",
 *   display_label = "Vipps MobilePay",
 *    forms = {
 *     "offsite-payment" = "Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect\RecurringLandingPageRedirectForm",
 *   },
 *   payment_method_types = {"vipps_mobilepay_wallet"},
 *   requires_billing_information = FALSE,
 * )
 */
class Recurring extends VippsMobilePayPaymentGatewayBase implements SupportsAuthorizationsInterface, SupportsRefundsInterface, SupportsStoredPaymentMethodsInterface {

  /**
   * Vipps manager factory.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface
   */
  protected $vippsManagerFactory;

  /**
   * The lock service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create(
      $container,
      $configuration,
      $plugin_id,
      $plugin_definition
    );
    $instance->vippsManagerFactory = $container->get('vipps_mobilepay.manager_factory');
    $instance->lock = $container->get('lock');
    $instance->logger = $container->get('logger.channel.vipps_mobilepay_commerce');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $agreement_reference = $order->getData('vipps_current_agreement_id');
    $charge_reference = $order->getData('vipps_current_charge_id');

    $paymentGatewayId = $this->parentEntity->id();
    $lockId = $paymentGatewayId . '__' . $agreement_reference;

    // Keep checking if lock could be acquired for 5 seconds, then start over.
    while ($this->lock->wait($lockId, 5) || !$this->lock->acquire($lockId)) {
      // Looks like lock cannot be acquired, hold for one second and retry.
      sleep(1);
      $this->logger->notice('Waiting for lock @lock to be released', ['@lock' => $lockId]);
    }
    $this->logger->notice('Lock @lock was acquired', ['@lock' => $lockId]);

    $matching_payments = $this->entityTypeManager->getStorage('commerce_payment')->loadByProperties([
      'remote_id' => $agreement_reference,
      'order_id' => $order->id(),
    ]);
    if (count($matching_payments) > 0) {
      $this->logger->debug('Payment :reference found already, skipping.', [':reference' => $agreement_reference]);
      return;
    }

    $payment_method_storage = $this->entityTypeManager->getStorage('commerce_payment_method');
    assert($payment_method_storage instanceof PaymentMethodStorageInterface);
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    assert($payment_storage instanceof PaymentStorageInterface);

    $settings = $this->configuration;
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
    $client = $this->vippsManagerFactory->get($connection);
    $response = (new RecurringPayment($client))->getAgreement($agreement_reference);

    $payment_method = $payment_method_storage->createForCustomer(
      'vipps_mobilepay_wallet',
      $this->parentEntity->id(),
      $order->getCustomerId(),
      $order->getBillingProfile());
    assert($payment_method instanceof PaymentMethodInterface);
    $payment_method->setRemoteId($agreement_reference);
    if ($response->getStart()?->getTimestamp()) {
      $payment_method->setCreatedTime($response->getStart()->getTimestamp());
    }
    if ($response->getStop()?->getTimestamp()) {
      $payment_method->setExpiresTime($response->getStop()->getTimestamp());
    }
    // @todo Validate status.
    $state = $response->getStatus();
    switch ($state) {
      case 'ACTIVE':
        $payment_method->setReusable(TRUE);
        $payment_method->setExpiresTime(0);
        $payment_method->setDefault(TRUE);
        // @todo create active payment method
        break;

      case 'STOPPED':
      case 'EXPIRED':
        // @todo create inactive payment method.
        throw new PaymentGatewayException('Subscription creation was aborted.');

      // Status CREATED means that final status has not yet been issued, hence
      // order must be kept locked in payment return state and status check must
      // be retried every 5s.
      case 'PENDING':
      default:
        sleep(5);
        $this->logger->error('Incorrect status @status, waiting for status change', ['@status' => $state]);
        throw new NeedsRedirectException(Url::fromRoute('<current>')->toString());
    }

    try {
      $payment_method->set('vipps_mobilepay_country_code', $response->getCountryCode());
      $payment_method->set('vipps_mobilepay_agreement_name', $response->getProductName());
      $payment_method->set('vipps_mobilepay_agreement_description', $response->getProductDescription());
      $payment_method->set('vipps_mobilepay_sub', $response->getSub());
      $payment_method->set('vipps_mobilepay_price', $this->minorUnitsConverter->fromMinorUnits($response->getPricing()->getAmount(), $response->getPricing()->getCurrency()));
      $response = (new UserInfo($client))->userInfo($response->getSub());
      $payment_method->set('vipps_mobilepay_telephone', $response->getPhoneNumber());
    }
    catch (\Exception $exception) {
      // @todo Queue for update in case something fails.
    }

    $response = (new RecurringPayment($client))->getCharge($agreement_reference, $charge_reference);
    $status = $response->getStatus();
    switch ($status) {
      case 'RESERVED':
        $state = 'authorization';
        break;

      case 'PARTIALLY_CAPTURED':
      case 'CHARGED':
        $state = 'completed';
        break;

      case 'CANCELLED':
        $state = 'authorization_voided';
        break;

      case 'PARTIALLY_REFUNDED':
        $state = 'partially_refunded';
        $refunded_amount = $this->minorUnitsConverter->fromMinorUnits($response->getSummary()->getRefunded(), $response->getCurrency());
        break;

      case 'REFUNDED':
        $state = 'refunded';
        break;

      case 'FAILED':
        throw new PaymentGatewayException('Charge failed at the payment gateway');

      case 'DUE':
      case 'PENDING':
      case 'PROCESSING':
      default:
        sleep(5);
        $this->logger->error('Incorrect status @status, waiting for status change', ['@status' => $state]);
        throw new NeedsRedirectException(Url::fromRoute('<current>')->toString());
    }

    // Save payment method.
    $payment_method->save();

    $payment_amount = $this->minorUnitsConverter->fromMinorUnits($response->getAmount(), $response->getCurrency());
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $payment_storage->create([
      'payment_method' => $payment_method->id(),
      'state' => $state,
      // Use the amount we get from remote, rather than the order total, since
      // it might be different.
      'amount' => $payment_amount,
      'payment_gateway' => $this->parentEntity->id(),
      'order_id' => $order->id(),
      'test' => $this->getMode() === 'test',
      'remote_id' => $charge_reference,
      'remote_state' => $response->getStatus(),
    ]);
    if (isset($refunded_amount)) {
      $payment->setRefundedAmount($refunded_amount);
    }
    $payment->save();
    $order->set('payment_method', $payment_method->id());
  }

  /**
   * Vipps treats onReturn and onCancel in the same way.
   *
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    static::onReturn($order, $request);
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    // Assert things.
    $this->assertPaymentState($payment, ['authorization']);
    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();

    // Adds support for partial captures.
    if ($amount->lessThan($payment->getAmount())) {
      /** @var \Drupal\commerce_payment\Entity\PaymentInterface $parent_payment */
      $parent_payment = $payment;
      $payment = $parent_payment->createDuplicate();
    }

    $charge_id = $payment->getRemoteId();
    $agreement_id = $payment->getPaymentMethod()->getRemoteId();

    $number = $this->minorUnitsConverter->toMinorUnits($amount);
    $request = new RequestCaptureCharge();
    $request->setAmount($number);
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new RecurringPayment($vipps);
      $payment_manager->captureCharge($agreement_id, $charge_id, $request);
    }
    catch (VippsException $exception) {
      throw new DeclineException($exception->getMessage());
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $payment->setState('completed');
    $payment->setAmount($amount);
    $payment->save();

    // Update parent payment if one exists.
    if (isset($parent_payment)) {
      $parent_payment->setAmount($parent_payment->getAmount()->subtract($amount));
      if ($parent_payment->getAmount()->isZero()) {
        $parent_payment->setState('authorization_voided');
      }
      $parent_payment->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);
    $charge_id = $payment->getRemoteId();
    $agreement_id = $payment->getPaymentMethod()->getRemoteId();

    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new RecurringPayment($vipps);
      $payment_manager->cancelCharge($agreement_id, $charge_id);
    }
    catch (VippsException $exception) {
      throw new DeclineException($exception->getMessage());
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $payment->setState('authorization_voided');
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    // Validate.
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);

    // Let's do some refunds.
    parent::assertRefundAmount($payment, $amount);

    $charge_id = $payment->getRemoteId();
    $agreement_id = $payment->getPaymentMethod()->getRemoteId();

    $number = $this->minorUnitsConverter->toMinorUnits($amount);
    $request = new RequestRefundCharge();
    $request->setAmount($number);
    $request->setDescription($this->t('Refund initiated by @store', ['@store' => $payment->getOrder()->getStore()->label()]));
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new RecurringPayment($vipps);
      $payment_manager->refundCharge($agreement_id, $charge_id, $request);
    }
    catch (VippsException $exception) {
      throw new DeclineException($exception->getMessage());
    }
    catch (\Exception $exception) {
      throw new DeclineException($exception->getMessage());
    }

    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();

  }

  /**
   * {@inheritdoc}
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE) {
    $price = $payment->getPaymentMethod()->vipps_mobilepay_price->first()->toPrice();
    assert($price instanceof Price);
    if ($price->getCurrencyCode() !== $payment->getAmount()->getCurrencyCode()) {
      throw new DeclineException('Currency mismatch');
    }
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')
        ->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new RecurringPayment($vipps);
      $request = new RequestCreateCharge();
      $request->setType('UNSCHEDULED');
      $request->setAmount($this->minorUnitsConverter->toMinorUnits($payment->getAmount()));
      $request->setDescription($this->t('Payment for order @order_id',
        ['@order_id' => $payment->getOrderId()]));
      if ($capture) {
        $request->setTransactionType('DIRECT_CAPTURE');
      }
      else {
        $request->setTransactionType('RESERVE_CAPTURE');
      }
      $response = $payment_manager->createCharge($payment->getPaymentMethod()
        ->getRemoteId(), $request);
    }
    catch (\Exception $exception) {
      // @todo Check if can recover.
      throw new SoftDeclineException($exception->getMessage());
    }
    $next_state = $capture ? 'completed' : 'authorization';
    $payment->setState($next_state);
    $payment->setRemoteId($response->getChargeId());
    // @todo Find out how long an authorization is valid, set its expiration.
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method) {
    try {
      $settings = $this->configuration;
      $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')
        ->load($settings['connection_profile']);
      $vipps = $this->vippsManagerFactory->get($connection);
      $payment_manager = new RecurringPayment($vipps);
      $request = new RequestUpdateAgreement();
      $request->setStatus('STOPPED');
      $payment_manager->updateAgreement($payment_method->getRemoteId(), $request);
    }
    catch (\Exception $exception) {
      // @todo Check if can recover.
      throw new SoftDeclineException($exception->getMessage());
    }
    $payment_method->delete();
  }

}
