<?php

namespace Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect;

use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TempStore\SharedTempStoreFactory;
use Drupal\Core\Url;
use Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface;
use Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use zaporylie\Vipps\Api\v3\Checkout;
use zaporylie\Vipps\Model\Checkout\v3\Amount as CheckoutAmount;
use zaporylie\Vipps\Model\Checkout\v3\Configuration;
use zaporylie\Vipps\Model\Checkout\v3\CreateCheckoutSessionRequest;
use zaporylie\Vipps\Model\Checkout\v3\MerchantInfo;
use zaporylie\Vipps\Model\Checkout\v3\OrderBottomLine;
use zaporylie\Vipps\Model\Checkout\v3\OrderLine;
use zaporylie\Vipps\Model\Checkout\v3\OrderSummary;
use zaporylie\Vipps\Model\Checkout\v3\Transaction;
use zaporylie\Vipps\Model\Checkout\v3\UnitInfo;

/**
 * Class CheckoutLandingPageRedirectForm.
 *
 * Handles the payment IFrame.
 */
class CheckoutLandingPageRedirectForm extends BasePaymentOffsiteForm implements ContainerInjectionInterface {

  /**
   * Used for query name and tempStore collection.
   *
   * @internal
   *   For internal use only!
   */
  const QUERY_NAME = 'vipps_mobilepay_commerce_payment_reference';
  const QUERY_NAME_ORDER_REFERENCE = 'vipps_mobilepay_commerce_payment_reference';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Vipps manager.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactory
   */
  protected $managerFactory;

  /**
   * The Chain Reference resolver.
   *
   * @var \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface
   */
  protected $chainReferenceResolver;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Shared temporary storage.
   *
   * @var \Drupal\Core\TempStore\SharedTempStore
   */
  protected $tempStore;

  /**
   * The minor units converter.
   *
   * @var \Drupal\commerce_price\MinorUnitsConverterInterface
   */
  protected $minorUnitsConverter;

  /**
   * VippsLandingPageRedirectForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface $manager_factory
   *   The vipps manager.
   * @param \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface $chain_reference_resolver
   *   The chain order id resolver.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\TempStore\SharedTempStoreFactory $tempStore
   *   The shared temporary storage factory.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $minor_units_converter
   *   The minor units converter.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, VippsMobilePayManagerFactoryInterface $manager_factory, ChainReferenceResolverInterface $chain_reference_resolver, EventDispatcherInterface $event_dispatcher, SharedTempStoreFactory $tempStore, MinorUnitsConverterInterface $minor_units_converter) {
    $this->entityTypeManager = $entity_type_manager;
    $this->managerFactory = $manager_factory;
    $this->chainReferenceResolver = $chain_reference_resolver;
    $this->eventDispatcher = $event_dispatcher;
    $this->tempStore = $tempStore->get(self::QUERY_NAME);
    $this->minorUnitsConverter = $minor_units_converter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('vipps_mobilepay.manager_factory'),
      $container->get('vipps_mobilepay_commerce.chain_reference_resolver'),
      $container->get('event_dispatcher'),
      $container->get('tempstore.shared'),
      $container->get('commerce_price.minor_units_converter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\Payment $payment */
    // When dumping here, we have a new entity, use that by default.
    $payment = $this->entity;
    /** @var \Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway\Checkout $plugin */
    $plugin = $payment->getPaymentGateway()->getPlugin();
    $settings = $payment->getPaymentGateway()->getPluginConfiguration();
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);

    // Generate payment ID.
    $payment->setRemoteId($settings['prefix'] . $this->chainReferenceResolver->resolve($payment));

    $order = $payment->getOrder();

    $order_summary = new OrderSummary();
    $order_summary->setOrderBottomLine((new OrderBottomLine())
      ->setCurrency($order->getTotalPrice()->getCurrencyCode()));

    $order_lines = [];
    foreach ($order->getItems() as $item) {
      $order_line = new OrderLine();
      $order_line->setName($item->label());
      $order_line->setId($item->getPurchasedEntityId());
      $order_line->setProductUrl($item->getPurchasedEntity()->toUrl('canonical')->setAbsolute()->toString());
      $order_line->setTotalAmount($this->minorUnitsConverter->toMinorUnits($item->getTotalPrice()));

      // The fallback value when no tax adjustment is found.
      $order_line->setTotalTaxAmount(0);
      $order_line->setTaxRate(0);
      $order_line->setTotalAmountExcludingTax($this->minorUnitsConverter->toMinorUnits($item->getTotalPrice()));

      // Only pass included tax adjustments (i.e "VAT"), non-included tax
      // adjustments are passed separately (i.e "Sales tax").
      $tax_adjustments = $item->getAdjustments(['tax']);
      if ($tax_adjustments && $tax_adjustments[0]->isIncluded()) {
        $tax_rate = $tax_adjustments[0]->getPercentage();
        $order_line->setTaxRate((int) Calculator::multiply($tax_rate, '10000'));
        $order_line->setTotalTaxAmount($this->minorUnitsConverter->toMinorUnits($tax_adjustments[0]->getAmount()));
        $order_line->setTotalAmountExcludingTax($this->minorUnitsConverter->toMinorUnits($item->getTotalPrice()->subtract($tax_adjustments[0]->getAmount())));
      }

      $order_line->setUnitInfo((new UnitInfo())->setQuantity($item->getQuantity()));

      $promotion_adjustments = $item->getAdjustments(['promotion']);
      foreach ($promotion_adjustments as $adjustment) {
        if (!$adjustment->isIncluded()) {
          continue;
        }
        if (!isset($promotion_sum)) {
          $promotion_sum = $adjustment->getAmount();
        }
        else {
          $promotion_sum->add($adjustment->getAmount());
        }
      }
      if (isset($promotion_sum)) {
        $order_line->setDiscount($this->minorUnitsConverter->toMinorUnits($promotion_sum));
      }
      $order_lines[] = $order_line;
    }

    $order_summary->setOrderLines($order_lines);

    try {
      /** @var \Drupal\vipps_mobilepay\VippsMobilepayConnectionInterface $connection */
      $vipps = $this->managerFactory->get($connection);

      $request = new CreateCheckoutSessionRequest();
      $request
        ->setType('PAYMENT')
        ->setMerchantInfo((new MerchantInfo())
          ->setCallbackUrl($this->addQueryToUrl($plugin->getNotifyUrl()->toString(), [self::QUERY_NAME => $payment->getRemoteId()]))
          ->setReturnUrl($this->addQueryToUrl($form['#return_url'], [self::QUERY_NAME => $payment->getRemoteId()]))
          ->setTermsAndConditionsUrl(Url::fromUserInput($settings['terms_path'] ?? '/')->setAbsolute()->toString())
          ->setCallbackAuthorizationToken(uniqid()))
        ->setTransaction((new Transaction())
          ->setAmount((new CheckoutAmount())
            ->setCurrency($payment->getAmount()->getCurrencyCode())
            ->setValue($this->minorUnitsConverter->toMinorUnits($payment->getAmount())))
          ->setReference($payment->getRemoteId())
          ->setPaymentDescription($this->t('Payment for @order_id', ['@order_id' => $order->id()]))
          ->setOrderSummary($order_summary));

      $request->setConfiguration((new Configuration())
        ->setElements('PaymentOnly')
        ->setShowOrderSummary(TRUE));

      $response = (new Checkout($vipps))
        ->createCheckoutSession($request);

      $form['iframe'] = [
        '#type' => 'container',
        '#id' => 'vipps-checkout-frame-container',
      ];

      $form['#attached']['library'][] = $payment->getPaymentGateway()->getPlugin()->getJsLibrary();
      $form['#attached']['drupalSettings']['vipps_mobilepay_commerce']['checkoutFrontendUrl'] = $response->getCheckoutFrontendUrl();
      $form['#attached']['drupalSettings']['vipps_mobilepay_commerce']['token'] = $response->getToken();
      $form['#attached']['drupalSettings']['vipps_mobilepay_commerce']['poolingUrl'] = $response->getPoolingUrl();
      // @todo Add language configuration.
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }
    return $form;

  }

  /**
   * Adds custom parameter to string-formed url.
   *
   * @param string $url
   *   Absolute URL.
   * @param array $query
   *   Array of query arguments to be added.
   *
   * @return string
   *   Absolute URL with query arguments.
   */
  private function addQueryToUrl($url, array $query) {
    $parsedUrl = parse_url($url);
    if ($parsedUrl['path'] == NULL) {
      $url .= '/';
    }
    $separator = !isset($parsedUrl['query']) ? '?' : '&';
    return $url .= $separator . http_build_query($query);
  }

}
