<?php

namespace Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect;

use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TempStore\SharedTempStoreFactory;
use Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface;
use Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use zaporylie\Vipps\Api\v1\EPayment;
use zaporylie\Vipps\Model\EPayment\v1\Amount;
use zaporylie\Vipps\Model\EPayment\v1\CreatePaymentRequest;
use zaporylie\Vipps\Model\EPayment\v1\PaymentMethod;

/**
 * Class EPaymentLandingPageRedirectForm.
 *
 * Handles the initiation of payments.
 */
class EPaymentLandingPageRedirectForm extends BasePaymentOffsiteForm implements ContainerInjectionInterface {

  /**
   * Used for query name and tempStore collection.
   *
   * @internal
   *   For internal use only!
   */
  const QUERY_NAME = 'vipps_mobilepay_commerce_payment_reference';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Vipps manager.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactory
   */
  protected $managerFactory;

  /**
   * The Chain Reference resolver.
   *
   * @var \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface
   */
  protected $chainReferenceResolver;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Shared temporary storage.
   *
   * @var \Drupal\Core\TempStore\SharedTempStore
   */
  protected $tempStore;

  /**
   * VippsLandingPageRedirectForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface $manager_factory
   *   The vipps manager.
   * @param \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface $chain_reference_resolver
   *   The chain order id resolver.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\TempStore\SharedTempStoreFactory $tempStore
   *   The shared temporary storage factory.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, VippsMobilePayManagerFactoryInterface $manager_factory, ChainReferenceResolverInterface $chain_reference_resolver, EventDispatcherInterface $event_dispatcher, SharedTempStoreFactory $tempStore) {
    $this->entityTypeManager = $entity_type_manager;
    $this->managerFactory = $manager_factory;
    $this->chainReferenceResolver = $chain_reference_resolver;
    $this->eventDispatcher = $event_dispatcher;
    $this->tempStore = $tempStore->get(self::QUERY_NAME);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('vipps_mobilepay.manager_factory'),
      $container->get('vipps_mobilepay_commerce.chain_reference_resolver'),
      $container->get('event_dispatcher'),
      $container->get('tempstore.shared')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\Payment $payment */
    // When dumping here, we have a new entity, use that by default.
    $payment = $this->entity;
    $settings = $payment->getPaymentGateway()->getPluginConfiguration();
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);

    // Create payment.
    $payment->setRemoteId($settings['prefix'] . $this->chainReferenceResolver->resolve($payment));

    // Save order.
    $order = $payment->getOrder();
    $order_changed = FALSE;

    if ($order->getData('vipps_current_transaction') !== $payment->getRemoteId()) {
      $order->setData('vipps_current_transaction', $payment->getRemoteId());
      $order_changed = TRUE;
    }

    try {
      /** @var \Drupal\vipps_mobilepay\VippsMobilepayConnectionInterface $connection */
      $vipps = $this->managerFactory->get($connection);

      $request = new CreatePaymentRequest();
      $request
        ->setAmount((new Amount())->setValue($payment->getAmount()->multiply(100)->getNumber())->setCurrency($payment->getAmount()->getCurrencyCode()))
        ->setPaymentMethod((new PaymentMethod())->setType('WALLET'))
        ->setReference($payment->getRemoteId())
        ->setUserFlow('WEB_REDIRECT')
        ->setReturnUrl($this->addQueryToUrl($form['#return_url'], [self::QUERY_NAME => $payment->getRemoteId()]))
        ->setMetadata(['order_id' => $order->id()]);

      $response = (new EPayment($vipps))->createPayment($request, base64_encode($payment->getRemoteId()));
      $url = $response->getRedirectUrl();

    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }

    // If the payment was successfully created at remote host.
    if ($order_changed === TRUE) {
      $order->save();
    }

    return $this->buildRedirectForm($form, $form_state, $url, []);
  }

  /**
   * Adds custom parameter to string-formed url.
   *
   * @param string $url
   *   Absolute URL.
   * @param array $query
   *   Array of query arguments to be added.
   *
   * @return string
   *   Absolute URL with query arguments.
   */
  private function addQueryToUrl($url, array $query) {
    $parsedUrl = parse_url($url);
    if ($parsedUrl['path'] == NULL) {
      $url .= '/';
    }
    $separator = !isset($parsedUrl['query']) ? '?' : '&';
    return $url .= $separator . http_build_query($query);
  }

}
