<?php

namespace Drupal\vipps_mobilepay_commerce\PluginForm\OffsiteRedirect;

use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TempStore\SharedTempStoreFactory;
use Drupal\Core\Url;
use Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface;
use Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use zaporylie\Vipps\Api\v3\RecurringPayment;
use zaporylie\Vipps\Model\RecurringPayment\v3\InitialCharge;
use zaporylie\Vipps\Model\RecurringPayment\v3\Pricing;
use zaporylie\Vipps\Model\RecurringPayment\v3\RequestCreateAgreement;

/**
 * Class CheckoutForm.
 *
 * Handles the initiation of payments.
 */
class RecurringLandingPageRedirectForm extends BasePaymentOffsiteForm implements ContainerInjectionInterface {

  /**
   * Used for query name and tempStore collection.
   *
   * @internal
   *   For internal use only!
   */
  const QUERY_AGREEMENT_PARAM = 'vipps_mobilepay_commerce_agreement_reference';

  /**
   * Used for query name and tempStore collection.
   *
   * @internal
   *   For internal use only!
   */
  const QUERY_CHARGE_NAME = 'vipps_mobilepay_commerce_payment_reference';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Vipps manager.
   *
   * @var \Drupal\vipps_mobilepay\VippsMobilePayManagerFactory
   */
  protected $managerFactory;

  /**
   * The Chain Reference resolver.
   *
   * @var \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface
   */
  protected $chainReferenceResolver;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Shared temporary storage.
   *
   * @var \Drupal\Core\TempStore\SharedTempStore
   */
  protected $tempStore;

  /**
   * The minor units converter.
   *
   * @var \Drupal\commerce_price\MinorUnitsConverterInterface
   */
  protected $minorUnitsConverter;

  /**
   * VippsLandingPageRedirectForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\vipps_mobilepay\VippsMobilePayManagerFactoryInterface $manager_factory
   *   The vipps manager.
   * @param \Drupal\vipps_mobilepay_commerce\Resolver\ChainReferenceResolverInterface $chain_reference_resolver
   *   The chain order id resolver.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\TempStore\SharedTempStoreFactory $tempStore
   *   The shared temporary storage factory.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $minor_units_converter
   *   The minor units converter.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, VippsMobilePayManagerFactoryInterface $manager_factory, ChainReferenceResolverInterface $chain_reference_resolver, EventDispatcherInterface $event_dispatcher, SharedTempStoreFactory $tempStore, MinorUnitsConverterInterface $minor_units_converter) {
    $this->entityTypeManager = $entity_type_manager;
    $this->managerFactory = $manager_factory;
    $this->chainReferenceResolver = $chain_reference_resolver;
    $this->eventDispatcher = $event_dispatcher;
    $this->tempStore = $tempStore->get(self::QUERY_AGREEMENT_PARAM);
    $this->minorUnitsConverter = $minor_units_converter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('vipps_mobilepay.manager_factory'),
      $container->get('vipps_mobilepay_commerce.chain_reference_resolver'),
      $container->get('event_dispatcher'),
      $container->get('tempstore.shared'),
      $container->get('commerce_price.minor_units_converter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\Payment $payment */
    // When dumping here, we have a new entity, use that by default.
    $payment = $this->entity;
    /** @var \Drupal\vipps_mobilepay_commerce\Plugin\Commerce\PaymentGateway\EPayment $plugin */
    $plugin = $payment->getPaymentGateway()->getPlugin();
    $settings = $payment->getPaymentGateway()->getPluginConfiguration();
    $connection = $this->entityTypeManager->getStorage('vipps_mobilepay_connection')->load($settings['connection_profile']);

    // Create payment.
    $payment->setRemoteId($settings['prefix'] . $this->chainReferenceResolver->resolve($payment));

    // Save order.
    $order = $payment->getOrder();
    $order_changed = FALSE;

    try {
      /** @var \Drupal\vipps_mobilepay\VippsMobilepayConnectionInterface $connection */
      $vipps = $this->managerFactory->get($connection);

      $request = new RequestCreateAgreement();
      // Phone number is required to be set on Payment Method.
      $request->setScope('phoneNumber');

      if (!$payment->getAmount()->isZero()) {
        $request->setInitialCharge((new InitialCharge())
          ->setAmount($this->minorUnitsConverter->toMinorUnits($payment->getAmount()))
          ->setTransactionType('RESERVE_CAPTURE')
          ->setOrderId($payment->getRemoteId())
          ->setDescription($this->t('Payment for @order_id', ['@order_id' => $order->id()])));
      }

      $request
        ->setPricing((new Pricing())
          ->setAmount($this->minorUnitsConverter->toMinorUnits($payment->getAmount()))
          ->setCurrency($payment->getAmount()->getCurrencyCode()));

      $request
        ->setMerchantRedirectUrl($this->addQueryToUrl($form['#return_url'], [self::QUERY_AGREEMENT_PARAM => $payment->getRemoteId()]))
        ->setMerchantAgreementUrl(Url::fromRoute('entity.commerce_payment_method.collection', ['user' => $order->getCustomer()->id()])->setAbsolute()->toString())
        ->setProductName($this->t('Payment method at @store', ['@store' => $order->getStore()->getName()]));

      $response = (new RecurringPayment($vipps))->createAgreement($request, base64_encode($payment->getRemoteId()));
      if ($response->getChargeId()) {
        $order->setData('vipps_current_charge_id', $response->getChargeId());
        $order_changed = TRUE;
      }
      if ($response->getAgreementId()) {
        $order->setData('vipps_current_agreement_id', $response->getAgreementId());
        $order_changed = TRUE;
      }

      $url = $response->getVippsConfirmationUrl();
    }
    catch (\Exception $exception) {
      throw new PaymentGatewayException($exception->getMessage());
    }

    // If the payment was successfully created at remote host.
    if ($order_changed === TRUE) {
      $order->save();
    }

    return $this->buildRedirectForm($form, $form_state, $url, []);
  }

  /**
   * Adds custom parameter to string-formed url.
   *
   * @param string $url
   *   Absolute URL.
   * @param array $query
   *   Array of query arguments to be added.
   *
   * @return string
   *   Absolute URL with query arguments.
   */
  private function addQueryToUrl($url, array $query) {
    $parsedUrl = parse_url($url);
    if ($parsedUrl['path'] == NULL) {
      $url .= '/';
    }
    $separator = !isset($parsedUrl['query']) ? '?' : '&';
    return $url .= $separator . http_build_query($query);
  }

}
